#include <node.h>
#include "node_wrapper_led_interface.hpp"

namespace node_wrappers {

using v8::Function;
using v8::FunctionCallbackInfo;
using v8::FunctionTemplate;
using v8::Isolate;
using v8::Local;
using v8::Number;
using v8::Object;
using v8::Persistent;
using v8::String;
using v8::Value;
using v8::Exception;

Persistent<Function> LedInterface::constructor;


void LedInterface::Init(Local<Object> exports) {
    Isolate* isolate = exports->GetIsolate();

    // Prepare constructor template
    Local<FunctionTemplate> tpl = FunctionTemplate::New(isolate, New);
    tpl->SetClassName(String::NewFromUtf8(isolate, "LedInterface"));
    tpl->InstanceTemplate()->SetInternalFieldCount(1);

    // Prototype
    NODE_SET_PROTOTYPE_METHOD(tpl, "check_connection", check_connection);
    NODE_SET_PROTOTYPE_METHOD(tpl, "set_luminance", set_luminance);
    NODE_SET_PROTOTYPE_METHOD(tpl, "get_luminance", get_luminance);
    NODE_SET_PROTOTYPE_METHOD(tpl, "set_fade", set_fade);
    NODE_SET_PROTOTYPE_METHOD(tpl, "get_u", get_u);
    NODE_SET_PROTOTYPE_METHOD(tpl, "get_i", get_i);

    constructor.Reset(isolate, tpl->GetFunction());
    exports->Set(String::NewFromUtf8(isolate, "LedInterface"), tpl->GetFunction());
}


void LedInterface::New(const FunctionCallbackInfo<Value>& args) {
    Isolate* isolate = args.GetIsolate();

    if (!args.IsConstructCall()) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "call with new")));
        return;
    }

    // https://code.google.com/archive/p/v8-juice/wikis/CommentaryOnV8.wiki#No_GC_guaranty
    try {
		LedInterface* obj = new LedInterface();
		obj->Wrap(args.This());
		
		args.GetReturnValue().Set(args.This());
	} catch (std::exception & e) {
        isolate->ThrowException(Exception::Error(String::NewFromUtf8(isolate, e.what())));
        return;
    }
}


void LedInterface::check_connection(const FunctionCallbackInfo<Value>& args) {
    Isolate* isolate = args.GetIsolate();

    if (args.Length() != 0) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong arguments")));
        return;
    }

    LedInterface* self = ObjectWrap::Unwrap<LedInterface>(args.Holder());
    try {
        self->li.check_connection();
    } catch (std::exception & e) {
        isolate->ThrowException(Exception::Error(String::NewFromUtf8(isolate, e.what())));
        return;
    }
}

void LedInterface::get_luminance(const FunctionCallbackInfo<Value>& args) {
    Isolate* isolate = args.GetIsolate();

    if (args.Length() != 0) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong arguments")));
        return;
    }

    LedInterface* self = ObjectWrap::Unwrap<LedInterface>(args.Holder());
    try {
        double ret = self->li.get_luminance();
        args.GetReturnValue().Set(Number::New(isolate, ret));
    } catch (std::exception & e) {
        isolate->ThrowException(Exception::Error(String::NewFromUtf8(isolate, e.what())));
        return;
    }
}


void LedInterface::set_luminance(const FunctionCallbackInfo<Value>& args) {
    Isolate* isolate = args.GetIsolate();

    if (args.Length() != 1 && !args[0]->IsNumber()) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong arguments")));
        return;
    }

    LedInterface* self = ObjectWrap::Unwrap<LedInterface>(args.Holder());
    try {
        self->li.set_luminance(args[0]->NumberValue());
    } catch (std::exception & e) {
        isolate->ThrowException(Exception::Error(String::NewFromUtf8(isolate, e.what())));
        return;
    }
}


void LedInterface::set_fade(const FunctionCallbackInfo<Value>& args) {
    Isolate* isolate = args.GetIsolate();

    if (args.Length() != 1 && !args[0]->IsNumber()) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong arguments")));
        return;
    }

    LedInterface* self = ObjectWrap::Unwrap<LedInterface>(args.Holder());
    try {
        self->li.set_fade(args[0]->NumberValue());
    } catch (std::exception & e) {
        isolate->ThrowException(Exception::Error(String::NewFromUtf8(isolate, e.what())));
        return;
    }
}


void LedInterface::get_i(const FunctionCallbackInfo<Value>& args) {
    Isolate* isolate = args.GetIsolate();

    if (args.Length() != 1 && !args[0]->IsNumber()) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong arguments")));
        return;
    }

    try {
        LedInterface* self = ObjectWrap::Unwrap<LedInterface>(args.Holder());
        double ret = self->li.get_i(args[0]->NumberValue());
        args.GetReturnValue().Set(Number::New(isolate, ret));
    } catch (std::exception & e) {
        isolate->ThrowException(Exception::Error(String::NewFromUtf8(isolate, e.what())));
        return;
    }
}


void LedInterface::get_u(const FunctionCallbackInfo<Value>& args) {
    Isolate* isolate = args.GetIsolate();

    if (args.Length() != 1 && !args[0]->IsNumber()) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong arguments")));
        return;
    }

    try {
        LedInterface* self = ObjectWrap::Unwrap<LedInterface>(args.Holder());
        double ret = self->li.get_u(args[0]->NumberValue());
        args.GetReturnValue().Set(Number::New(isolate, ret));
    } catch (std::exception & e) {
        isolate->ThrowException(Exception::Error(String::NewFromUtf8(isolate, e.what())));
        return;
    }
}


}
