#include <node.h>
#include "node_wrapper_yitran_nc.hpp"
#include "yitran_packet.hpp"
#include <iostream>
#include <iomanip>

namespace node_wrappers {

using v8::Function;
using v8::FunctionCallbackInfo;
using v8::FunctionTemplate;
using v8::Isolate;
using v8::Local;
using v8::Handle;
using v8::Array;
using v8::Number;
using v8::Integer;
using v8::Uint32;
using v8::Boolean;
using v8::Object;
using v8::Persistent;
using v8::String;
using v8::Value;
using v8::Exception;
using v8::Symbol;

Persistent<Function> YitranNc::constructor;

void YitranNc::Init(Local<Object> exports) {
    Isolate* isolate = exports->GetIsolate();

    // Prepare constructor template
    Local<FunctionTemplate> tpl = FunctionTemplate::New(isolate, New);
    tpl->SetClassName(String::NewFromUtf8(isolate, "YitranNc"));
    tpl->InstanceTemplate()->SetInternalFieldCount(1);

    // Prototype
    NODE_SET_PROTOTYPE_METHOD(tpl, "reset_and_go_online", reset_and_go_online);
    // From yitran_host
    NODE_SET_PROTOTYPE_METHOD(tpl, "wait_for_runtime_indication", wait_for_runtime_indication);
    NODE_SET_PROTOTYPE_METHOD(tpl, "non_acknowledged_packet_tx", non_acknowledged_packet_tx);
    NODE_SET_PROTOTYPE_METHOD(tpl, "decode_rx_intranetworking_packet_indication", decode_rx_intranetworking_packet_indication);
    NODE_SET_PROTOTYPE_METHOD(tpl, "acknowledged_packet_tx", acknowledged_packet_tx);

    NODE_SET_PROTOTYPE_METHOD(tpl, "send_data_exchange_request", send_data_exchange_request);
    NODE_SET_PROTOTYPE_METHOD(tpl, "decode_data_exchange_reply", decode_data_exchange_reply);
    
    tpl->InstanceTemplate()->SetAccessor( String::NewFromUtf8(isolate, "recently_received_packet"), GetRecentlyReceivedPacket, SetRecentlyReceivedPacket);

    constructor.Reset(isolate, tpl->GetFunction());
    exports->Set(String::NewFromUtf8(isolate, "YitranNc"), tpl->GetFunction());
}


void YitranNc::New(const FunctionCallbackInfo<Value>& args) {
    Isolate* isolate = args.GetIsolate();

    if (!args.IsConstructCall()) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "call with new")));
        return;
    }

    if (args.Length() != 1 && !args[0]->IsBoolean()) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "First argument is required and must be bool!")));
        return;
    }

    // https://code.google.com/archive/p/v8-juice/wikis/CommentaryOnV8.wiki#No_GC_guaranty
    try {
		YitranNc* obj = new YitranNc(args[0]->BooleanValue() );
		obj->Wrap(args.This());
		
		args.GetReturnValue().Set(args.This());
	} catch (std::exception & e) {
        isolate->ThrowException(Exception::Error(String::NewFromUtf8(isolate, e.what())));
        return;
    }
}

void YitranNc::reset_and_go_online(const FunctionCallbackInfo<Value>& args) {
    Isolate* isolate = args.GetIsolate();

    if (args.Length() != 1 && !args[0]->IsNumber()) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong arguments")));
        return;
    }

    YitranNc* self = ObjectWrap::Unwrap<YitranNc>(args.Holder());
    try {
        bool ret = self->nc.reset_and_go_online(args[0]->Uint32Value());
        args.GetReturnValue().Set(Boolean::New(isolate, ret));
    } catch (std::exception & e) {
        isolate->ThrowException(Exception::Error(String::NewFromUtf8(isolate, e.what())));
        return;
    }
}

void YitranNc::wait_for_runtime_indication(const FunctionCallbackInfo<Value>& args) {
    Isolate* isolate = args.GetIsolate();

    if (args.Length() != 1 && !args[0]->IsNumber()) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong arguments")));
        return;
    }

    YitranNc* self = ObjectWrap::Unwrap<YitranNc>(args.Holder());
    try {
        yitran_host::runtime_indication ret = self->nc.wait_for_runtime_indication(args[0]->NumberValue()); /*TODO:Int32Value*/
        args.GetReturnValue().Set(Integer::New(isolate, ret));
    } catch (std::exception & e) {
        isolate->ThrowException(Exception::Error(String::NewFromUtf8(isolate, e.what())));
        return;
    }
}

void YitranNc::non_acknowledged_packet_tx(const FunctionCallbackInfo<Value>& args) {
    Isolate* isolate = args.GetIsolate();

    if (args.Length() != 2 && !args[0]->IsUint32() && !args[1]->IsArray() ) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong arguments")));
        return;
    }

    YitranNc* self = ObjectWrap::Unwrap<YitranNc>(args.Holder());
    try {
		Handle<Array> jsArray = Handle<Array>::Cast(args[1]);
		unsigned int length = jsArray->Length();
		unsigned char *message = new unsigned char[jsArray->Length()];
		
		for (unsigned int i = 0; i < length; i++) {
			message[i] = jsArray->Get(i)->Uint32Value() ;
		}		

        self->nc.non_acknowledged_packet_tx(args[0]->Uint32Value(), message, length);
        delete[] message;
	} catch (std::exception & e) {
        isolate->ThrowException(Exception::Error(String::NewFromUtf8(isolate, e.what())));
        return;
    }
}

void YitranNc::decode_rx_intranetworking_packet_indication(const FunctionCallbackInfo<Value>& args) {
    Isolate* isolate = args.GetIsolate();

    if (args.Length() != 0) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong arguments")));
        return;
    }

    YitranNc* self = ObjectWrap::Unwrap<YitranNc>(args.Holder());
    try {
        self->nc.decode_rx_intranetworking_packet_indication();
    } catch (std::exception & e) {
        isolate->ThrowException(Exception::Error(String::NewFromUtf8(isolate, e.what())));
        return;
    }
}

void YitranNc::acknowledged_packet_tx(const FunctionCallbackInfo<Value>& args) {
    Isolate* isolate = args.GetIsolate();

    if (args.Length() != 3 && !args[0]->IsUint32() && !args[1]->IsArray() && !args[2]->IsUint32() ) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong arguments")));
        return;
    }

    YitranNc* self = ObjectWrap::Unwrap<YitranNc>(args.Holder());
    try {
		Handle<Array> jsArray = Handle<Array>::Cast(args[1]);
		unsigned int length = jsArray->Length();
		unsigned char *message = new unsigned char[length];
				
		for (unsigned int i = 0; i < length; i++) {
			message[i] = jsArray->Get(i)->Uint32Value() ;
		}
		
		
        bool ret = self->nc.acknowledged_packet_tx(args[0]->Uint32Value(), message, length, args[2]->Uint32Value() );
        args.GetReturnValue().Set(Boolean::New(isolate, ret));
        delete[] message;
    } catch (std::exception & e) {
        isolate->ThrowException(Exception::Error(String::NewFromUtf8(isolate, e.what())));
        return;
    }
}

void YitranNc::GetRecentlyReceivedPacket(v8::Local< v8::String > property, const v8::PropertyCallbackInfo< Value > &info)
{
	Isolate* isolate = info.GetIsolate();
	Local<v8::Context> context = isolate->GetCurrent()->GetCurrentContext();
	
    YitranNc* self = ObjectWrap::Unwrap<YitranNc>(info.Holder());
    yitran_host::yitran_packet recently_received_packet = self->nc.recently_received_packet;

	Local<Object> obj = Object::New(isolate);
	
	obj->Set( context , String::NewFromUtf8(isolate, "command_start"), Uint32::New(isolate, recently_received_packet.command_start) );
	obj->Set( context , String::NewFromUtf8(isolate, "payload_size"), Uint32::New(isolate, recently_received_packet.payload_size) );
	obj->Set( context , String::NewFromUtf8(isolate, "type"), Uint32::New(isolate, recently_received_packet.type) );
	obj->Set( context , String::NewFromUtf8(isolate, "opcode"), Uint32::New(isolate, recently_received_packet.opcode) );
	
	Local<Array> payload = Array::New(isolate, 256);
	for (unsigned i = 0; i < 256; ++i ) {
		payload->Set(i, Uint32::New(isolate,recently_received_packet.payload[i]));
	}
	obj->Set( context , String::NewFromUtf8(isolate, "payload"), payload );
		
	obj->Set( context , String::NewFromUtf8(isolate, "origin_id"), Uint32::New(isolate, recently_received_packet.origin_id) );
	obj->Set( context , String::NewFromUtf8(isolate, "final_id"), Uint32::New(isolate, recently_received_packet.final_id) );
	obj->Set( context , String::NewFromUtf8(isolate, "user_payload_size"), Uint32::New(isolate, recently_received_packet.user_payload_size) );

	Local<Array> user_payload = Array::New(isolate, 256);
	for (unsigned i = 0; i < 256; ++i ) {
		user_payload->Set(i, Uint32::New(isolate,recently_received_packet.user_payload[i]));
	}	
	obj->Set( context , String::NewFromUtf8(isolate, "user_payload"), user_payload );
	
	info.GetReturnValue().Set( obj ); 
}

void YitranNc::SetRecentlyReceivedPacket(v8::Local< v8::String > property, v8::Local< v8::Value > value, const v8::PropertyCallbackInfo< void > &info)
{
	Isolate* isolate = info.GetIsolate();
	//Local<v8::Context> context = isolate->GetCurrent()->GetCurrentContext();
	
    //YitranNc* self = ObjectWrap::Unwrap<YitranNc>(info.Holder());
    //yitran_host::yitran_packet recently_received_packet;

	isolate->ThrowException(Exception::Error(String::NewFromUtf8(isolate, "*** WARNING ***: SetRecentlyReceivedPacket() not implemented in node_wrapper_yitran_nc.cpp.")));
	
}

void YitranNc::send_data_exchange_request(const FunctionCallbackInfo<Value>& args)
{
    Isolate* isolate = args.GetIsolate();

    if (args.Length() != 2 && !args[0]->IsUint32() && !args[1]->IsUint32()) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong arguments")));
        return;
    }

    YitranNc* self = ObjectWrap::Unwrap<YitranNc>(args.Holder());
    try {
        self->nc.send_data_exchange_request(args[0]->Uint32Value(), args[1]->Uint32Value());
    } catch (std::exception & e) {
        isolate->ThrowException(Exception::Error(String::NewFromUtf8(isolate, e.what())));
        return;
    }

}

void YitranNc::decode_data_exchange_reply(const FunctionCallbackInfo<Value>& args)
{
    Isolate* isolate = args.GetIsolate();
    Local<v8::Context> context = isolate->GetCurrent()->GetCurrentContext();

    if (args.Length() != 0) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong arguments")));
        return;
    }

    YitranNc* self = ObjectWrap::Unwrap<YitranNc>(args.Holder());
    try {
        yitran_host::rs_hk_data_set hk_data_set = self->nc.decode_data_exchange_reply(); /*TODO:Int32Value*/
       	
       	Local<Object> obj = Object::New(isolate);
	
		obj->Set( context , String::NewFromUtf8(isolate, "u1"), Number::New(isolate, hk_data_set.u1) );
		obj->Set( context , String::NewFromUtf8(isolate, "u2"), Number::New(isolate, hk_data_set.u2) );
		obj->Set( context , String::NewFromUtf8(isolate, "u3"), Number::New(isolate, hk_data_set.u3) );
		obj->Set( context , String::NewFromUtf8(isolate, "u4"), Number::New(isolate, hk_data_set.u4) );

		obj->Set( context , String::NewFromUtf8(isolate, "i1"), Number::New(isolate, hk_data_set.i1) );
		obj->Set( context , String::NewFromUtf8(isolate, "i2"), Number::New(isolate, hk_data_set.i2) );
		obj->Set( context , String::NewFromUtf8(isolate, "i3"), Number::New(isolate, hk_data_set.i3) );
		obj->Set( context , String::NewFromUtf8(isolate, "i4"), Number::New(isolate, hk_data_set.i4) );

		obj->Set( context , String::NewFromUtf8(isolate, "temperature"), Number::New(isolate, hk_data_set.temperature) );
		obj->Set( context , String::NewFromUtf8(isolate, "humidity"), Number::New(isolate, hk_data_set.humidity) );

		obj->Set( context , String::NewFromUtf8(isolate, "luminance"), Uint32::New(isolate, hk_data_set.luminance) );
		obj->Set( context , String::NewFromUtf8(isolate, "pylon_id"), Uint32::New(isolate, hk_data_set.pylon_id) );
        
        args.GetReturnValue().Set(obj);
        
    } catch (std::exception & e) {
        isolate->ThrowException(Exception::Error(String::NewFromUtf8(isolate, e.what())));
        return;
    }

}


}
