#include <node.h>
#include "node_wrapper_yitran_rs.hpp"

namespace node_wrappers {

using v8::Function;
using v8::FunctionCallbackInfo;
using v8::FunctionTemplate;
using v8::Isolate;
using v8::Local;
using v8::Number;
using v8::Handle;
using v8::Array;
using v8::Integer;
using v8::Boolean;
using v8::Object;
using v8::Persistent;
using v8::String;
using v8::Value;
using v8::Exception;

Persistent<Function> YitranRs::constructor;

void YitranRs::Init(Local<Object> exports) {
    Isolate* isolate = exports->GetIsolate();

    // Prepare constructor template
    Local<FunctionTemplate> tpl = FunctionTemplate::New(isolate, New);
    tpl->SetClassName(String::NewFromUtf8(isolate, "YitranRs"));
    tpl->InstanceTemplate()->SetInternalFieldCount(1);

    // Prototype
    NODE_SET_PROTOTYPE_METHOD(tpl, "reset_and_go_online", reset_and_go_online);
    // From yitran_host
    NODE_SET_PROTOTYPE_METHOD(tpl, "wait_for_runtime_indication", wait_for_runtime_indication);
    NODE_SET_PROTOTYPE_METHOD(tpl, "non_acknowledged_packet_tx", non_acknowledged_packet_tx);
    NODE_SET_PROTOTYPE_METHOD(tpl, "decode_rx_intranetworking_packet_indication", decode_rx_intranetworking_packet_indication);
    NODE_SET_PROTOTYPE_METHOD(tpl, "acknowledged_packet_tx", acknowledged_packet_tx);

    constructor.Reset(isolate, tpl->GetFunction());
    exports->Set(String::NewFromUtf8(isolate, "YitranRs"), tpl->GetFunction());
}


void YitranRs::New(const FunctionCallbackInfo<Value>& args) {
    Isolate* isolate = args.GetIsolate();

    if (!args.IsConstructCall()) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "call with new")));
        return;
    }

    if (args.Length() != 1 && !args[0]->IsBoolean()) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "First argument is required and must be bool!")));
        return;
    }

    // https://code.google.com/archive/p/v8-juice/wikis/CommentaryOnV8.wiki#No_GC_guaranty
    try {
		YitranRs* obj = new YitranRs(args[0]->BooleanValue() );
		obj->Wrap(args.This());
		
		args.GetReturnValue().Set(args.This());
	} catch (std::exception & e) {
        isolate->ThrowException(Exception::Error(String::NewFromUtf8(isolate, e.what())));
        return;
    }
}

void YitranRs::reset_and_go_online(const FunctionCallbackInfo<Value>& args) {
    Isolate* isolate = args.GetIsolate();

    if (args.Length() != 1 && !args[0]->IsNumber()) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong arguments")));
        return;
    }

    YitranRs* self = ObjectWrap::Unwrap<YitranRs>(args.Holder());
    try {
        bool ret = self->rs.reset_and_go_online(args[0]->Uint32Value());
        args.GetReturnValue().Set(Boolean::New(isolate, ret));
    } catch (std::exception & e) {
        isolate->ThrowException(Exception::Error(String::NewFromUtf8(isolate, e.what())));
        return;
    }
}

void YitranRs::wait_for_runtime_indication(const FunctionCallbackInfo<Value>& args) {
    Isolate* isolate = args.GetIsolate();

    if (args.Length() != 1 && !args[0]->IsNumber()) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong arguments")));
        return;
    }

    YitranRs* self = ObjectWrap::Unwrap<YitranRs>(args.Holder());
    try {
        yitran_host::runtime_indication ret = self->rs.wait_for_runtime_indication(args[0]->NumberValue()); /*TODO:Int32Value*/
        args.GetReturnValue().Set(Integer::New(isolate, ret));
    } catch (std::exception & e) {
        isolate->ThrowException(Exception::Error(String::NewFromUtf8(isolate, e.what())));
        return;
    }
}

void YitranRs::non_acknowledged_packet_tx(const FunctionCallbackInfo<Value>& args) {
    Isolate* isolate = args.GetIsolate();

    if (args.Length() != 2 && !args[0]->IsUint32() && !args[1]->IsArray() ) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong arguments")));
        return;
    }

    YitranRs* self = ObjectWrap::Unwrap<YitranRs>(args.Holder());
    try {
		Handle<Array> jsArray = Handle<Array>::Cast(args[1]);
		unsigned int length = jsArray->Length();
		unsigned char *message = new unsigned char[jsArray->Length()];
		
		for (unsigned int i = 0; i < length; i++) {
			message[i] = jsArray->Get(i)->Uint32Value() ;
		}		
        self->rs.non_acknowledged_packet_tx(args[0]->Uint32Value(), message, length);
        delete[] message;
    } catch (std::exception & e) {
        isolate->ThrowException(Exception::Error(String::NewFromUtf8(isolate, e.what())));
        return;
    }
}

void YitranRs::decode_rx_intranetworking_packet_indication(const FunctionCallbackInfo<Value>& args) {
    Isolate* isolate = args.GetIsolate();

    if (args.Length() != 0) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong arguments")));
        return;
    }

    YitranRs* self = ObjectWrap::Unwrap<YitranRs>(args.Holder());
    try {
        self->rs.decode_rx_intranetworking_packet_indication();
    } catch (std::exception & e) {
        isolate->ThrowException(Exception::Error(String::NewFromUtf8(isolate, e.what())));
        return;
    }
}

void YitranRs::acknowledged_packet_tx(const FunctionCallbackInfo<Value>& args) {
    Isolate* isolate = args.GetIsolate();

    if (args.Length() != 3 && !args[0]->IsUint32() && !args[1]->IsArray() && !args[2]->IsUint32() ) {
        isolate->ThrowException(Exception::TypeError(String::NewFromUtf8(isolate, "Wrong arguments")));
        return;
    }

    YitranRs* self = ObjectWrap::Unwrap<YitranRs>(args.Holder());
    try {
		Handle<Array> jsArray = Handle<Array>::Cast(args[1]);
		unsigned int length = jsArray->Length();
		unsigned char *message = new unsigned char[length];
		
		for (unsigned int i = 0; i < length; i++) {
			message[i] = jsArray->Get(i)->Uint32Value() ;
		}		

        bool ret = self->rs.acknowledged_packet_tx(args[0]->Uint32Value(), message, length, args[2]->Uint32Value());
        args.GetReturnValue().Set(Boolean::New(isolate, ret));
        delete[] message;        
    } catch (std::exception & e) {
        isolate->ThrowException(Exception::Error(String::NewFromUtf8(isolate, e.what())));
        return;
    }
}

}
