#include <iostream>
#include <unistd.h>
#include <time.h>
#include <vector>
#include <string>
#include "mraa.hpp"
#include "yitran_host.hpp"
#include "yitran_nc.hpp"
#include "yitran_node_descriptor.hpp"
#include "json.hpp"

#include "mqtt/client.h"
#include "persistence.h"

using json = nlohmann::json;

const std::string SERVER_ADDRESS { "tcp://localhost:1883" };
const std::string CLIENT_ID { "simplified_nc" };
const std::string TOPIC { "rshk" };
const int QOS = 0;

// TODO: export to its own .h
class user_callback : public virtual mqtt::callback
{
	void connection_lost(const std::string& cause) override {
		std::cout << "\nConnection lost" << std::endl;
		if (!cause.empty())
			std::cout << "\tcause: " << cause << std::endl;
	}

	void delivery_complete(mqtt::delivery_token_ptr tok) override {
		std::cout << "\n\t[Delivery complete for token: "
			<< (tok ? tok->get_message_id() : -1) << "]" << std::endl;
	}

public:
};

//-------------------------------------------------------------------------------------------------
int application_simplified_nc (std::vector<unsigned> const & pylon_ids) {

	yitran_host::yitran_nc nc(false);

	yitran_host::rs_hk_data_set recent_rs_hk_data_set;
	unsigned char luminance = 1;
	yitran_host::runtime_indication recent_runtime_indication = yitran_host::unknown_indication;

	unsigned char pylon_pointer = 0;

	std::cout << "Initialize NC..." << std::endl;
	nc.reset_and_go_online(2);

	sample_mem_persistence persist;
	mqtt::client client(SERVER_ADDRESS, CLIENT_ID, &persist);

	user_callback cb;
	client.set_callback(cb);

	mqtt::connect_options connOpts;
	connOpts.set_keep_alive_interval(20);
	connOpts.set_clean_session(true);


	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	try {
		std::cout << "\nConnecting to MQTT Broker..." << std::endl;
		client.connect(connOpts);
		std::cout << "...MQTT Connection OK" << std::endl;
	}
	catch (const mqtt::exception& exc) {
		std::cerr << exc.what() << std::endl;
		return 1;
	}

	while (1) {

		if ( luminance == 1 ) luminance = 255;
		else luminance = 1;

		// Send data exchange request.
		try {
			nc.send_data_exchange_request(pylon_ids[pylon_pointer], luminance);
			std::cout << "Data exchange request sent to " << pylon_ids[pylon_pointer] << "." << std::endl;
		} catch (std::exception& e) {
			std::cout << e.what() << std::endl;
			// TODO: adapt to use vector
			// if ( pylon_pointer == 1 ) pylon_pointer = 0;
			// else ++pylon_pointer;
			continue;
		}

		//---------------------------------------------------------

		// Wait for the data exchange reply.
		while (1) {
			try {
				recent_runtime_indication = nc.wait_for_runtime_indication(10000);
				if ( recent_runtime_indication == yitran_host::rx_intranetworking_packet ) {
					nc.decode_rx_intranetworking_packet_indication();
					try {
						recent_rs_hk_data_set = nc.decode_data_exchange_reply();
					} catch (std::exception& e) {
						std::cout << e.what() << std::endl;
						break;
					}
					if ( recent_rs_hk_data_set.pylon_id == pylon_ids[pylon_pointer] ) {
						json mqttMessage;
						mqttMessage["pylon_id"] = recent_rs_hk_data_set.pylon_id;
						mqttMessage["voltages"] = {recent_rs_hk_data_set.u1, recent_rs_hk_data_set.u2, recent_rs_hk_data_set.u3, recent_rs_hk_data_set.u4};
						mqttMessage["currents"] = {recent_rs_hk_data_set.i1, recent_rs_hk_data_set.i2, recent_rs_hk_data_set.i3, recent_rs_hk_data_set.i4};
						mqttMessage["luminance"] = recent_rs_hk_data_set.luminance;
						mqttMessage["temp"] = recent_rs_hk_data_set.temperature;
						mqttMessage["rh"] = recent_rs_hk_data_set.humidity;

						auto pubmsg = mqtt::make_message(TOPIC, mqttMessage.dump());
						pubmsg->set_qos(QOS);
						client.publish(pubmsg);

						std::cout << "---------------------------------------" << std::endl;
						std::cout << "RS #" << pylon_ids[pylon_pointer] << " HK data set: " << std::endl;
						std::cout << "Pylon ID: " << recent_rs_hk_data_set.pylon_id << std::endl;
						std::cout << "LED string #1: " << recent_rs_hk_data_set.u1 << " V, " << recent_rs_hk_data_set.i1 << " mA" << std::endl;
						std::cout << "LED string #2: " << recent_rs_hk_data_set.u2 << " V, " << recent_rs_hk_data_set.i2 << " mA" << std::endl;
						std::cout << "LED string #3: " << recent_rs_hk_data_set.u3 << " V, " << recent_rs_hk_data_set.i3 << " mA" << std::endl;
						std::cout << "LED string #4: " << recent_rs_hk_data_set.u4 << " V, " << recent_rs_hk_data_set.i4 << " mA" << std::endl;
						std::cout << "Temperature: " << recent_rs_hk_data_set.temperature << " Celsius" << std::endl;
						std::cout << "Humidity: " << recent_rs_hk_data_set.humidity << "%" << std::endl;
						std::cout << "Luminance: " << 100 * recent_rs_hk_data_set.luminance / 255 << "%" << std::endl;
						std::cout << "---------------------------------------" << std::endl;
						break;
					} else continue;

				} else continue;
			} 
			catch (const mqtt::persistence_exception& exc) {
				std::cerr << "Persistence Error: " << exc.what() << " ["
					<< exc.get_reason_code() << "]" << std::endl;
				return 1;
			}
			catch (const mqtt::exception& exc) {
				std::cerr << exc.what() << std::endl;
				return 1;
			}			
			catch (std::exception& e) {
				std::cout << e.what() << std::endl;
				break;
			}
		}

		//---------------------------------------------------------

		sleep(1);
		// if ( pylon_pointer == 1 ) {
		// 	pylon_pointer = 0;
			// if ( luminance == 0 ) luminance = 255;
			// else luminance = 0;
		// }
		// else ++pylon_pointer;

	}

	try {
		// Disconnect
		std::cout << "\nDisconnecting from MQTT broker..." << std::endl;
		client.disconnect();
		std::cout << "...OK" << std::endl;
	}
	catch (const mqtt::exception& exc) {
		std::cerr << exc.what() << std::endl;
		return 1;
	}			
	
	return 0;

}
//-------------------------------------------------------------------------------------------------
