#include "yitran_host.hpp"
#include "stdio.h"
#include <iostream>
#include <iomanip>

//-------------------------------------------------------------------------------------------------
namespace yitran_host {

	yitran_host::yitran_host(bool use_console_port) {
		yitran_reset_pin = new mraa::Gpio(48);	// GPIO15
		yitran_reset_pin->dir(mraa::DIR_OUT);
		if ( use_console_port == false ) yitran_uart = new mraa::Uart(0);
		else yitran_uart = new mraa::Uart("/dev/ttyMFD2");
		yitran_uart->setBaudRate(38400);
		yitran_uart->setMode(8, mraa::UART_PARITY_NONE, 1);
		yitran_uart->setFlowcontrol(false, false);
	}



	yitran_host::~yitran_host() {
		if ( yitran_reset_pin != NULL ) delete yitran_reset_pin;
		if ( yitran_uart != NULL ) delete yitran_uart;
	}


	void yitran_host::reset() {
		yitran_reset_pin->write(1);
		sleep(1);
		yitran_reset_pin->write(0);
	}



	void yitran_host::send_packet(const yitran_packet& packet_2_send) {

		unsigned char checksum = 0;

		// send command start field
		yitran_uart->write((char*)(&packet_2_send.command_start), 1);

		// send length field
		unsigned length = packet_2_send.payload_size + 2;
		char length_H = (0x0000FF00 & length) >> 8;
		char length_L = (0x000000FF & length);
		yitran_uart->write(&length_L, 1);
		yitran_uart->write(&length_H, 1);
		checksum += length_L;
		checksum += length_H;

		// send type field
		yitran_uart->write((char*)(&packet_2_send.type) ,1);
		checksum += packet_2_send.type;

		// send opcode field
		yitran_uart->write((char*)(&packet_2_send.opcode) ,1);
		checksum += packet_2_send.opcode;

		// send payload
		for ( unsigned i = 0; i < packet_2_send.payload_size; ++i ) {
			yitran_uart->write((char*)(&packet_2_send.payload[i]) ,1);
			checksum += packet_2_send.payload[i];
		}

		// send checksum
		yitran_uart->write((char*)(&checksum) ,1);

	}



	void yitran_host::get_packet(unsigned timeout_ms) {

		char buffer = 0;
		unsigned char calculated_checksum = 0;

		// look for command start: 0xCA or 0xCC (in case of BSP welcome message)
		while (1) {
			if ( get_byte(timeout_ms, &buffer) ) {

				if ( (0x000000FF & buffer) != 0xca && (0x000000FF & buffer) != 0xcc ) {
					// packets before command start symbol, are discarded
					printf("Byte discarded: %x\n", (0x000000FF & buffer));
					continue;
				} else {
					//printf("Byte received: %x\n", (0x000000FF & buffer));
					recently_received_packet.command_start = buffer;
					break;
				}

			} else {
				// get_byte timed out...
				throw(packet_reception_timeout_exception());
			}
		}

		// receive packet length field: 2 bytes
		char length_buffer[2];
		length_buffer[0] = 0;
		length_buffer[1] = 0;
		for ( int i = 0; i < 2; ++i ) {
			if ( get_byte(timeout_ms, &buffer) ) {
				length_buffer[i] = buffer;
				calculated_checksum += buffer;
			} else {
				// get_byte timed out...
				throw(packet_reception_timeout_exception());
			}
		}

		// calculate length field
		recently_received_packet.payload_size = (((0x000000FF & length_buffer[1]) << 8) | (0x000000FF & length_buffer[0])) - 2;

		// receive packet type: 1 byte
		if ( get_byte(timeout_ms, &buffer) ) {
			recently_received_packet.type = buffer;
			calculated_checksum += buffer;
		} else {
			// get_byte timed out...
			throw(packet_reception_timeout_exception());
		}

		// receive opcode: 1 byte
		if ( get_byte(timeout_ms, &buffer) ) {
			recently_received_packet.opcode = buffer;
			calculated_checksum += buffer;
		} else {
			// get_byte timed out...
			throw(packet_reception_timeout_exception());
		}

		// receive packet payload
		// the number of received bytes is determined by the packet length field: receive_packet_buffer[0, 1]
		for ( unsigned i = 0; i < recently_received_packet.payload_size; ++i ) {
			if ( get_byte(timeout_ms, &buffer) ) {
				recently_received_packet.payload[i] = buffer;
				calculated_checksum += buffer;
			} else {
				// get_byte timed out...
				throw(packet_reception_timeout_exception());
			}
		}

		// receive checksum: 1 byte
		if ( get_byte(timeout_ms, &buffer) ) {
			if ( calculated_checksum != (unsigned char)(buffer) ) {
				// checksum error
				throw(checksum_error_exception());
			}
		} else {
			// get_byte timed out...
			throw(packet_reception_timeout_exception());
		}

		// the packet is complete

	}



	runtime_indication yitran_host::wait_for_runtime_indication(unsigned timeout_ms) {
		get_packet(timeout_ms);
		if ( recently_received_packet.command_start != 0xca ) return unknown_indication;

		switch (recently_received_packet.opcode) {
			case 0xb1: return connectivity_status_with_rs;
			case 0xbe: return new_connection_to_nc;
			case 0xbb: return disconnected_from_nc;
			case 0xba: return connected_to_nc;
			case 0x68: return rx_intranetworking_packet;
			default: return unknown_indication;
		}
	}



	void yitran_host::decode_rx_intranetworking_packet_indication() {
		recently_received_packet.origin_id = (recently_received_packet.payload[17] << 8) | recently_received_packet.payload[16];
		recently_received_packet.final_id = (recently_received_packet.payload[19] << 8) | recently_received_packet.payload[18];
		recently_received_packet.user_payload_size = recently_received_packet.payload_size - 21;
		for ( unsigned i = 0; i < recently_received_packet.payload_size-21; ++i ) {
			recently_received_packet.user_payload[i] = recently_received_packet.payload[i+21];
		}
	}



	bool yitran_host::wait_for_reset_response(unsigned timeout_ms) {
		get_packet(timeout_ms);
		if ( recently_received_packet.command_start == 0xca && recently_received_packet.opcode == 0x20 ) return true;
		else return false;
	}



	bool yitran_host::wait_for_go_online_response(unsigned timeout_ms) {
		get_packet(timeout_ms);
		if ( recently_received_packet.command_start == 0xca && recently_received_packet.opcode == 0x22 ) return true;
		else return false;
	}



	void yitran_host::acknowledged_nop(unsigned timeout_ms) {
		send_nop();
		get_packet(timeout_ms);
	}



	bool yitran_host::acknowledged_set_network_size(unsigned char size, unsigned timeout_ms) {
		set_network_size(size);
		get_packet(timeout_ms);
		if ( recently_received_packet.payload[0] == 0x00 ) return false;
		else return true;
	}



	bool yitran_host::verified_set_network_size(unsigned char size, unsigned timeout_ms) {
		if ( !acknowledged_set_network_size(size, timeout_ms) ) return false;
		if ( get_network_size(timeout_ms) != size ) return false;
		else return true;
	}



	bool yitran_host::acknowledged_set_node_id(short node_id, unsigned timeout_ms) {
		set_node_id(node_id);
		get_packet(timeout_ms);
		if ( recently_received_packet.payload[0] == 0x00 ) return false;
		else return true;
	}



	bool yitran_host::verified_set_node_id(short node_id, unsigned timeout_ms) {
		if ( !acknowledged_set_node_id(node_id, timeout_ms) ) return false;
		if ( get_node_id(timeout_ms) != node_id ) return false;
		else return true;
	}



	bool yitran_host::get_operation_mode(unsigned timeout_ms) {
		yitran_packet p;
		p.command_start = 0xca;
		p.payload_size = 5;
		p.type = 0x00;
		p.opcode = 0x42;
		p.payload[0] = 0x06; // table type
		p.payload[1] = 0x31; // index LSB
		p.payload[2] = 0x00; // index MSB
		p.payload[3] = 0x01; // number of parameters to read LSB
		p.payload[4] = 0x00; // number of parameters to read MSB
		send_packet(p);

		// wait for response packet
		get_packet(timeout_ms);

		if ( recently_received_packet.payload[1] == 0x00 ) return false;
		else return true;
	}



	unsigned char yitran_host::get_network_size(unsigned timeout_ms) {
		yitran_packet p;
		p.command_start = 0xca;
		p.payload_size = 5;
		p.type = 0x00;
		p.opcode = 0x42;
		p.payload[0] = 0x06; // table type
		p.payload[1] = 0x38; // index LSB
		p.payload[2] = 0x00; // index MSB
		p.payload[3] = 0x01; // number of parameters to read LSB
		p.payload[4] = 0x00; // number of parameters to read MSB
		send_packet(p);

		// wait for response packet
		get_packet(timeout_ms);

		return recently_received_packet.payload[1];
	}



	short yitran_host::get_node_id(unsigned timeout_ms) {
		yitran_packet p;
		p.command_start = 0xca;
		p.payload_size = 5;
		p.type = 0x00;
		p.opcode = 0x42;
		p.payload[0] = 0x06; // table type
		p.payload[1] = 0x19; // index LSB
		p.payload[2] = 0x00; // index MSB
		p.payload[3] = 0x01; // number of parameters to read LSB
		p.payload[4] = 0x00; // number of parameters to read MSB
		send_packet(p);

		// wait for response packet
		get_packet(timeout_ms);

		return (recently_received_packet.payload[2] << 8) | recently_received_packet.payload[1];
	}



	bool yitran_host::acknowledged_save_device_parameters(unsigned timeout_ms) {
		save_device_parameters();
		get_packet(timeout_ms);
		if ( recently_received_packet.payload[0] == 0x00 ) return false;
		else return true;
	}



	bool yitran_host::acknowledged_go_online(unsigned timeout_ms) {
		go_online();
		get_packet(timeout_ms);
		if ( recently_received_packet.payload[0] == 0x00 ) return false;
		else return true;
	}



	bool yitran_host::acknowledged_packet_tx(short destination_id, unsigned char* user_payload, unsigned char user_payload_size, unsigned timeout_ms) {

		// Check user payload size
		if ( user_payload_size > 245 ) throw(invalid_user_payload_size_exception());

		yitran_packet p;
		p.command_start = 0xca;
		p.payload_size = user_payload_size + 11;
		p.type = 0x00;
		p.opcode = 0x60;

		// Host TX Header
			// TX flags
			p.payload[0] = 0x01; // Data Service Type: Intranetworking Unicast
			p.payload[1] = 0x00; // Priority: Normal
			p.payload[2] = 0x00; // Ack Service: No Ack required
			p.payload[3] = 0x08; // Hops: default
			p.payload[4] = 0x07; // Gain
			p.payload[5] = 0x01; // Tag LSB
			p.payload[6] = 0x00; // Tag MSB
			p.payload[7] = 0x00; // Encrypt

			// Destination port
			p.payload[8] = 0x00;

			// Target ID
			p.payload[9] = destination_id & 0x00FF; // target ID LSB
			p.payload[10] = (destination_id & 0xFF00) >> 8; // target ID MSB

		// user payload
		for ( unsigned i = 0; i < user_payload_size; ++i ) {
			p.payload[i+11] = user_payload[i];
		}

		send_packet(p);

		// wait for "Packet TX Admission" response
		while (1) {
			get_packet(timeout_ms);
			if ( recently_received_packet.command_start == 0xca && recently_received_packet.opcode == 0x60 ) {
				if ( recently_received_packet.payload[0] == 0x00 ) throw(packet_tx_admission_failed_exception());
				switch (recently_received_packet.payload[2]) {
					case 0x00:
						//printf("Packet TX Admission result: Accepted\n");
						break;
					case 0x01: printf("Packet TX Admission result: Rejected (no memory)\n"); return false;
					case 0x02: printf("Packet TX Admission result: Rejected (Fatal error)\n"); return false;
					case 0x03: printf("Packet TX Admission result: Rejected (node ID not found)\n"); return false;
					default: printf("Packet TX Admission result: INVALID"); return false;
				}
				break;
			}
			else {
				printf("Packet discarded while waiting for Packet TX Admission response\n");
				continue;
			}
		}

		// wait for "Packet TX Transmission" response
		while (1) {
			get_packet(timeout_ms);
			if ( recently_received_packet.command_start == 0xca && recently_received_packet.opcode == 0x60 ) {
				if ( recently_received_packet.payload[0] == 0x00 ) throw(packet_tx_admission_failed_exception());
				switch (recently_received_packet.payload[2]) {
					case 0x00:
						//printf("Packet TX Transmission result: Transmitted successfully\n");
						return true;
					case 0x01: printf("Packet TX Transmission result: N/A\n"); return false;
					case 0x02: printf("Packet TX Transmission result: No Acknowledge\n"); return false;
					case 0x03: printf("Packet TX Transmission result: No Resources (Target)\n"); return false;
					case 0x04: printf("Packet TX Transmission result: Blocked\n"); return false;
					case 0x05: printf("Packet TX Transmission result: Unknown error\n"); return false;
					default: printf("Packet TX Transmission result: INVALID\n"); return false;
				}
				break;
			}
			else {
				printf("Packet discarded while waiting for Packet TX Transmission response\n");
				continue;
			}
		}

	}

	void yitran_host::non_acknowledged_packet_tx(short destination_id, unsigned char* user_payload, unsigned char user_payload_size) {

		// Check user payload size
		if ( user_payload_size > 245 ) throw(invalid_user_payload_size_exception());

		yitran_packet p;
		p.command_start = 0xca;
		p.payload_size = user_payload_size + 11;
		p.type = 0x00;
		p.opcode = 0x60;

		// Host TX Header
			// TX flags
			p.payload[0] = 0x01; // Data Service Type: Intranetworking Unicast
			p.payload[1] = 0x00; // Priority: Normal
			p.payload[2] = 0x00; // Ack Service: No Ack required
			p.payload[3] = 0x08; // Hops: default
			p.payload[4] = 0x07; // Gain
			p.payload[5] = 0x01; // Tag LSB
			p.payload[6] = 0x00; // Tag MSB
			p.payload[7] = 0x00; // Encrypt

			// Destination port
			p.payload[8] = 0x00;

			// Target ID
			p.payload[9] = destination_id & 0x00FF; // target ID LSB
			p.payload[10] = (destination_id & 0xFF00) >> 8; // target ID MSB

		// user payload
		for ( unsigned i = 0; i < user_payload_size; ++i ) {
			p.payload[i+11] = user_payload[i];
		}

		send_packet(p);

	}

	void yitran_host::print_recently_received_packet() {
		printf("\nCommand start: %x\n", recently_received_packet.command_start);
		printf("Payload size: %d\n", recently_received_packet.payload_size);
		printf("Type: %x\n", recently_received_packet.type);
		printf("Opcode: %x\n", recently_received_packet.opcode);
		printf("Payload: ");
		for ( unsigned i = 0; i < recently_received_packet.payload_size; ++i ) {
			printf("%x, ", recently_received_packet.payload[i]);
		}
		printf("\n\n");
	}


	//---------------------------------------------------------------------------------------------
	// PRIVATE ------------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------------------
	bool yitran_host::get_byte(unsigned timeout_ms, char* buffer) {
		if ( yitran_uart->dataAvailable(timeout_ms) ) {
			yitran_uart->read(buffer, 1);
			return true;
		} else return false;
	}



	void yitran_host::send_nop() {
		yitran_packet p;
		p.command_start = 0xca;
		p.payload_size = 0;
		p.type = 0x00;
		p.opcode = 0x00;
		send_packet(p);
	}



	void yitran_host::set_nc_operation_mode() {
		yitran_packet p;
		p.command_start = 0xca;
		p.payload_size = 5;
		p.type = 0x00;
		p.opcode = 0x41;
		p.payload[0] = 0x06; // table type
		p.payload[1] = 0x31; // index LSB
		p.payload[2] = 0x00; // index MSB
		p.payload[3] = 0x03; // operation mode LSB
		p.payload[4] = 0x00; // operation mode MSB
		send_packet(p);
	}



	void yitran_host::set_rs_operation_mode() {
		yitran_packet p;
		p.command_start = 0xca;
		p.payload_size = 5;
		p.type = 0x00;
		p.opcode = 0x41;
		p.payload[0] = 0x06; // table type
		p.payload[1] = 0x31; // index LSB
		p.payload[2] = 0x00; // index MSB
		p.payload[3] = 0x00; // operation mode LSB
		p.payload[4] = 0x00; // operation mode MSB
		send_packet(p);
	}



	void yitran_host::set_nc_database_size(unsigned char size) {
		yitran_packet p;
		p.command_start = 0xca;
		p.payload_size = 5;
		p.type = 0x00;
		p.opcode = 0x41;
		p.payload[0] = 0x06; // table type
		p.payload[1] = 0x5b; // index LSB
		p.payload[2] = 0x00; // index MSB
		p.payload[3] = size;
		p.payload[4] = 0x00;
		send_packet(p);
	}



	void yitran_host::set_network_size(unsigned char size) {
		yitran_packet p;
		p.command_start = 0xca;
		p.payload_size = 5;
		p.type = 0x00;
		p.opcode = 0x41;
		p.payload[0] = 0x06; // table type
		p.payload[1] = 0x38; // index LSB
		p.payload[2] = 0x00; // index MSB
		p.payload[3] = size;
		p.payload[4] = 0x00;
		send_packet(p);
	}



	void yitran_host::set_node_id(short node_id) {
		yitran_packet p;
		p.command_start = 0xca;
		p.payload_size = 5;
		p.type = 0x00;
		p.opcode = 0x41;
		p.payload[0] = 0x06; // table type
		p.payload[1] = 0x19; // index LSB
		p.payload[2] = 0x00; // index MSB
		p.payload[3] = (0x00FF & node_id); // node ID LSB
		p.payload[4] = (0xFF00 & node_id) >> 8; // node ID MSB
		send_packet(p);
	}



	void yitran_host::save_device_parameters() {
		yitran_packet p;
		p.command_start = 0xca;
		p.payload_size = 1;
		p.type = 0x00;
		p.opcode = 0x43;
		p.payload[0] = 0xFF; // table type
		send_packet(p);
	}



	void yitran_host::go_online() {
		yitran_packet p;
		p.command_start = 0xca;
		p.payload_size = 0;
		p.type = 0x00;
		p.opcode = 0x22;
		send_packet(p);
	}

}
//-------------------------------------------------------------------------------------------------
