#ifndef _YITRAN_RS_
#define _YITRAN_RS_
//-------------------------------------------------------------------------------------------------
#include "mraa.hpp"
#include "yitran_packet.hpp"
#include "yitran_node_descriptor.hpp"
#include "yitran_host.hpp"
#include <exception>
//-------------------------------------------------------------------------------------------------
namespace yitran_host {

	class yitran_rs : public yitran_host {

		public:
			yitran_rs(bool use_console_port);
			~yitran_rs();


			/*
			 * The function initializes the Yitran modem as an RS node, saves the device parameters into the non-volatile memory, resets the modem and sends a go-online request.
			 */
			bool reset_and_go_online(unsigned network_size);



			/*
			 * The function initialized the RS node and after the address reception is done, it sends a login message to the NC.
			 * In case of a timeout, the function throws a packet_reception_timeout_exception.
			 * If the RS initialization, address reception, and login message transmission is successful,
			 * the function returns true, otherwise it returns false.
			 */
			bool initialize_and_login(unsigned char network_size);



			/*
			 * The function waits for the data traffic enabled indication from the NC.
			 * If the indication received is a valid enable data traffic indication, the function returns true,
			 * otherwise it returns false.
			 * In case of a timeout, the function throws a packet_reception_timeout_exception.
			 */
			bool wait_for_data_traffic_enabled_indication(unsigned timeout_ms);



			/*
			 * After receiving a "Disconnected from NC" indication, this function may be used to
			 * get the relevant information stored in the payload (command data) field of the indication.
			 * Relevant information: Reason of disconnection -> return value
			 */
			disconnect_reason decode_disconnected_from_nc_indication();



			/*
			 * After receiving a "Connected to NC" indication, this function may be used to
			 * get the relevant information stored in the payload (command data) field of the indication.
			 * Relevant information:
			 * 		- parent ID: The logical ID of the node's parent. -> nd.parent_id
			 */
			void decode_connected_to_nc_indication(yitran_node_descriptor& nd);



			/*
			 * The function receives a single packet. If the received packet is a "Connected to NC" indication,
			 * the function returns true, otherwise it returns false.
			 *
			 * Note: Connecting to NC timeout recommendation: 30 sec
			 */
			bool wait_for_connected_to_nc_indication(unsigned timeout_ms);



			/*
			 * The function sends a "Set Device Parameters" request setting the operation mode to RS.
			 * Thereafter, the function waits for the response and checks the "status" field.
			 * If no response comes back, the function throws a packet_reception_timeout_exception.
			 * Return value:
			 *   true, if the status field of the response indicates that the command was executed successfully
			 *   false, if the status field of the response indicates that the command execution failed
			 */
			bool acknowledged_set_rs_operation_mode(unsigned timeout_ms);



			/*
			 * The function sends a "Set Device Parameters" request setting the operation mode to RS.
			 * Thereafter, the function waits for the response and checks the "status" field.
			 * If no response comes back, the function throws a packet_reception_timeout_exception.
			 * If a response is received, the function checks the "status" field.
			 * If the "status" field indicates command error, the function returns false immediately.
			 * If the "status" field indicates a successful command, the function sends a "Get Device Parameters" request,
			 * reading the operation mode.
			 * Thereafter, the function waits for the response.
			 * If no response comes back, the function throws a packet_reception_timeout_exception.
			 * If a response is received, the function checks the operation mode value.
			 * If the operation mode is set to RS, the function returns true, otherwise it returns false.
			 */
			bool verified_set_rs_operation_mode(unsigned timeout_ms);




		private:
			//...

	};

}
//-------------------------------------------------------------------------------------------------
#endif // _YITRAN_RS_
//-------------------------------------------------------------------------------------------------
